import os
import tempfile
import warnings
from pathlib import Path
from uuid import uuid4

from typing_extensions import Any, Callable, Generic, TypeVar, get_args

__all__ = ["IndexWrapper", "load_index_from_temp_file", "save_index_to_temp_file"]

IS_WINDOWS = os.name == "nt"

T = TypeVar("T")  # Generic type for index wrapper


class IndexWrapper(Generic[T]):
    """A general purpose wrapper around indexing structures that require values observed from the data.

    For example, some indexing structures require an argument ``dim`` for the data dimensionality of the index,
    which might not be known at class instantiation time, but can automatically be inferred once data is available.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialization just stores the parameters.

        :param args: Any positional arguments to the underlying index.
        :param kwargs: Any keyword arguments to the underlying index.
        """
        super().__init__()
        if type(self) is IndexWrapper:
            msg = "Cannot instantiate 'IndexWrapper', subclass it with the correct generic index type."
            raise TypeError(msg)

        self.index: T | None = None  # might be set by subclass
        self.args = args
        self.kwargs = kwargs

    def __call__(self, *args: Any, **kwargs: Any) -> T:
        """Initialize the underlying index using the original ``args`` and ``kwargs`` with additional parameters.

        :param kwargs: Keyword arguments to be supplied to the underlying index.
        :return: The instantiated underlying index.
        """
        # ``__orig_bases__`` bases returns the generic base class, e.g. ``IndexWrapper[usearch.Index]`` (should be one)
        # ``get_args`` extract the generic type arguments from the base, e.g. ``usearch.Index`` (should be one)
        index = self.index if self.index is not None else get_args(self.__class__.__orig_bases__[0])[0]  # type: ignore[reportAttributeAccessIssue]
        return index(*args, *self.args, **self.kwargs, **kwargs)  # type: ignore[reportCallIssue]


def save_index_to_temp_file(save_fn: Callable[..., Any], **save_kwargs: Any) -> bytes:
    # there might be cleanup errors in multiprocessing setting on windows
    with tempfile.TemporaryDirectory(ignore_cleanup_errors=IS_WINDOWS) as tmp_dir:
        file_path = Path(tmp_dir) / uuid4().hex
        save_fn(str(file_path), **save_kwargs)
        return file_path.read_bytes()


def load_index_from_temp_file(index_bytes: bytes, load_fn: Callable[..., Any], **load_kwargs: Any) -> None:
    # cannot delete the file on windows, ``delete_on_close`` is only available with Python >= 3.12
    with tempfile.NamedTemporaryFile("wb", delete=not IS_WINDOWS) as file:
        file.write(index_bytes)
        file.flush()  # otherwise errors with 'Inappropriate ioctl for device'

        # it's not allowed to re-open the file on windows
        # https://stackoverflow.com/questions/23212435/permission-denied-to-write-to-my-temporary-file
        if IS_WINDOWS:
            file.close()

        # open the file and load the index
        load_fn(file.name, **load_kwargs)

        # manually delete the file on windows
        if IS_WINDOWS:
            try:
                file.close()
                Path(file.name).unlink(missing_ok=True)
            except PermissionError:
                msg = f"Could not delete file '{file.name}', the file might be used by another process."
                warnings.warn(msg, stacklevel=1)



from pathlib import Path

import numpy as np
from annoy import AnnoyIndex
from safecheck import Float, Int64, NumpyArray, Real, typecheck
from typing_extensions import Any, Iterable, Literal, overload

from ._base import NearestNeighbors
from ._base._helpers import load_index_from_temp_file, save_index_to_temp_file

__all__ = [
    "AnnoyNeighbors",
]


class AnnoyNeighbors(NearestNeighbors):
    available_metrics = Literal["angular", "euclidean", "manhattan", "hamming", "dot"]

    @typecheck
    def __init__(
        self,
        *,
        metric: available_metrics = "euclidean",
        n_trees: int = 32,
        n_search_neighbors: int | None = None,
        random_seed: int | None = None,
        disk_build_path: str | Path | None = None,
        save_index_path: str | Path | None = None,
        load_index_path: str | Path | None = None,
        load_index_dim: int | None = None,
        prefault: bool = False,
    ) -> None:
        """Instantiate annoy nearest neighbors.

        :param metric: One of ["angular", "euclidean", "manhattan", "hamming", "dot"]
        :param n_trees: Builds a forest of ``n_trees`` trees. More trees gives higher precision when querying.
        :param n_search_neighbors: Inspect up to ``n_search_neighbors`` nodes, default is ``n_search_neighbors`` * n.
        :param random_seed: Initialize the random number generator with the given seed.
        :param disk_build_path: Build the index on disk given the ``disk_build_path``.
        :param save_index_path: Save the index on disk after build given ``save_index_path``.
        :param load_index_path: Loads (mmaps) an index from disk given ``load_index_path``, requires ``load_index_dim``.
        :param load_index_dim: Specify the dimension for a loaded index.
        :param prefault: If prefault is set to True, it will pre-read the entire file into memory (mmap MAP_POPULATE).
        """
        super().__init__()
        self._index: AnnoyIndex | None = None

        # mmap the raw annoy index if ``load_index_path`` is given
        if (path := self.parameters.load_index_path) is not None:
            # raise an exception if ``load_index_dim`` is not given
            if self.parameters.load_index_dim is None:
                msg = f"Cannot load index from path '{path}' if 'load_index_dim' is None."
                raise AssertionError(msg)

            index = AnnoyIndex(self.parameters.load_index_dim, self.parameters.metric)
            index.load(str(path))
            self._index = index
            self.__fitted__ = True

    @overload
    def fit(self, data: Iterable[Real[NumpyArray, "d"]]) -> "AnnoyNeighbors": ...

    @overload
    def fit(self, data: Real[NumpyArray, "n d"]) -> "AnnoyNeighbors": ...

    def fit(self, data):  # type: ignore[reportGeneralTypeIssues]
        n_samples, n_dim = data.shape
        self.parameters.load_index_dim = n_dim
        index = AnnoyIndex(n_dim, metric=self.parameters.metric)

        if (seed := self.parameters.random_seed) is not None:
            index.set_seed(seed)

        if (path := self.parameters.disk_build_path) is not None:
            index.on_disk_build(str(path))

        for i in range(n_samples):
            index.add_item(i, data[i].tolist())

        index.build(self.parameters.n_trees)

        if (path := self.parameters.save_index_path) is not None:
            index.save(str(path), prefault=self.parameters.prefault)

        self._index = index
        return self

    def query(
        self,
        point: Float[NumpyArray, "d"],
        n_neighbors: int,
    ) -> tuple[Int64[NumpyArray, "{n_neighbors}"], Float[NumpyArray, "{n_neighbors}"]]:
        """No runtime type checks, because query is used in the ``query_batch`` loop."""
        idx, dist = self._index.get_nns_by_vector(
            point.tolist(),
            n_neighbors,
            self.parameters.n_search_neighbors,
            include_distances=True,
        )
        return np.array(idx, dtype=np.int64), np.array(dist, dtype=point.dtype)

    def query_idx(self, point: Float[NumpyArray, "d"], n_neighbors: int) -> Int64[NumpyArray, "{n_neighbors}"]:
        idx = self._index.get_nns_by_vector(
            point.tolist(),
            n_neighbors,
            self.parameters.n_search_neighbors,
        )
        return np.array(idx, dtype=np.int64)

    def __setstate__(self, state: dict[str, Any]) -> None:
        """Load the index using a temporary file."""
        super().__setstate__(state)
        if self._index is not None:
            index_bytes = state.pop("_index")
            self._index = AnnoyIndex(self.parameters.load_index_dim, self.parameters.metric)
            load_index_from_temp_file(index_bytes, load_fn=self._index.load, prefault=self.parameters.prefault)

    def __getstate__(self) -> dict[str, Any]:
        """Save the index to a temporary file and store the bytes of the file."""
        state = super().__getstate__()
        if self._index is not None:
            state["_index"] = save_index_to_temp_file(self._index.save, prefault=self.parameters.prefault)
        return state



import warnings
from pathlib import Path

import usearch.index as usearch
from numpy import atleast_2d  # numpy is a dependecy of usearch, import should be safe
from safecheck import Float, Integer, NumpyArray, typecheck
from typing_extensions import Any

from ._base import NearestNeighbors
from ._base._helpers import IndexWrapper, save_index_to_temp_file

__all__ = ["UsearchIndex", "UsearchNeighbors"]


class UsearchIndex(IndexWrapper[usearch.Index]): ...


DEFAULT_INDEX = UsearchIndex(metric="l2sq")


class UsearchNeighbors(NearestNeighbors):
    """A simple wrapper around ``usearch.Index``.

    References
    ----------
        - https://github.com/unum-cloud/usearch

    """

    @typecheck
    def __init__(
        self,
        *,
        index: usearch.Index | UsearchIndex = DEFAULT_INDEX,
        exact_search: bool = False,
        copy_data: bool = True,
        threads_fit: int = 0,
        threads_search: int = 0,
        log_fit: bool = False,
        log_search: bool = False,
        progress_fit: usearch.ProgressCallback | None = None,
        progress_search: usearch.ProgressCallback | None = None,
        progress_save: usearch.ProgressCallback | None = None,
        progress_load: usearch.ProgressCallback | None = None,
        save_index_path: str | Path | None = None,
        load_index_path: str | Path | None = None,
        map_file_index: bool = False,
        add_data_on_fit: bool = True,
    ) -> None:
        """Instantiate Usearch nearest neighbors.

        :param index: A UsearchIndex wrapped index or a usearch.Index.
        :param exact_search: Bypass index and use bruteforce exact search.
        :param copy_data: Should the index store a copy of vectors.
        :param threads_fit: Optimal number of cores to use for index creation.
        :param threads_search: Optimal number of cores to use for index search.
        :param log_fit: Whether to print the progress bar on index creation.
        :param log_search: Whether to print the progress bar on index search.
        :param progress_fit: Callback to report stats of the index creation progress.
        :param progress_search: Callback to report stats of the index search progress.
        :param progress_save: Callback to report stats of the index save progress.
        :param progress_load: Callback to report stats of the index load progress.
        :param save_index_path: Save the index after index creation.
        :param load_index_path: Load an existing index directly on ``__init__``.
        :param map_file_index: Memory map an index after creation or on load.
        :param add_data_on_fit: Add data to the index directly on index creation.
        """
        super().__init__()
        # to be defined in ``fit``
        self._index: usearch.Index | None = None

        # load the index from file
        if (path := self.parameters.load_index_path) is not None:
            if save_index_path is not None:
                msg = f"Using 'load_index_path' ({path}), but 'save_index_path' is not None and is ignored."
                warnings.warn(msg, stacklevel=1)

            if not isinstance(self.parameters.index, usearch.Index):
                msg = f"Using 'load_index_path' ({path}), but no index given, using 'restore' to load."
                warnings.warn(msg, stacklevel=1)
            else:
                # set the index here, such that load uses it to load from file
                self._index = self.parameters.index

            # load index and manually set ``__fitted__``, such that fit validation is not enabled
            self._load_index(path)
            self.__fitted__ = True

    @typecheck
    def fit(self, data: Float[NumpyArray, "n d"]) -> "UsearchNeighbors":
        _, dim = data.shape
        self._index = self._create_index(dim)

        if self.parameters.add_data_on_fit:
            # data might be added directly on fit, or using the ``add`` method
            self.add(data)

        if (path := self.parameters.save_index_path) is not None:
            self._index.save(
                path_or_buffer=path,
                progress=self.parameters.progress_save,
            )
            if self.parameters.map_file_index:
                self._index.reset()
                self._index.view(path)

        return self

    @typecheck
    def add(self, data: Float[NumpyArray, "n d"]) -> "UsearchNeighbors":
        self._index.add(
            keys=None,  # type: ignore[reportArgumentType]
            vectors=data,
            copy=self.parameters.copy_data,
            threads=self.parameters.threads_fit,
            log=self.parameters.log_fit,
            progress=self.parameters.progress_fit,
        )
        return self

    @typecheck
    def query(
        self,
        point: Float[NumpyArray, "d"],
        n_neighbors: int,
    ) -> tuple[Integer[NumpyArray, "{n_neighbors}"], Float[NumpyArray, "{n_neighbors}"]]:
        match = self._search(point, n_neighbors)
        return match.keys, match.distances

    @typecheck
    def query_batch(
        self,
        points: Float[NumpyArray, "m d"],
        n_neighbors: int,
    ) -> tuple[Integer[NumpyArray, "m {n_neighbors}"], Float[NumpyArray, "m {n_neighbors}"]]:
        match = self._search(points, n_neighbors)
        idxs, dists = match.keys, match.distances
        return tuple(atleast_2d(idxs, dists))  # type: ignore[reportReturnType]

    def _search(
        self,
        point_or_points: Float[NumpyArray, "d"] | Float[NumpyArray, "m d"],
        n_neighbors: int,
    ) -> usearch.Matches | usearch.BatchMatches:
        return self._index.search(
            point_or_points,
            count=n_neighbors,
            threads=self.parameters.threads_search,
            exact=self.parameters.exact_search,
            log=self.parameters.log_search,
            progress=self.parameters.progress_search,
        )

    def _create_index(self, dim: int) -> usearch.Index:
        index = self.parameters.index
        # if the index is an existing index, use it, otherwise the index should be a callable returning an index
        return index if isinstance(index, usearch.Index) else index(ndim=dim)

    def _load_index(self, path: str | Path) -> None:
        if not Path(path).is_file():
            msg = f"Tried to load index from '{path}', but no file fount at path."
            raise AssertionError(msg)

        if isinstance(self._index, usearch.Index):
            load = self._index.view if self.parameters.map_file_index else self._index.load
            load(path, progress=self.parameters.progress_load)
        else:
            self._index = usearch.Index.restore(path, view=self.parameters.map_file_index)

    def __setstate__(self, state: dict[str, Any]) -> None:
        """Load the index using a temporary file."""
        super().__setstate__(state)
        if self._index is not None:
            index_bytes = state.pop("_index")
            self._index = usearch.Index.restore(index_bytes, view=False)
            return

        if path := self.parameters.load_index_path:
            self._load_index(path)
            return

        if (path := self.parameters.save_index_path) and self.is_fitted:
            self._load_index(path)
            return

    def __getstate__(self) -> dict[str, Any]:
        """Save the index to a temporary file and store the bytes of the file."""
        state = super().__getstate__()
        if self._index is not None:

            # index path is given, don't persist the index data
            if self.parameters.load_index_path is not None or self.parameters.save_index_path is not None:
                state["_index"] = None  # load using restore, don't persist the index
                return state

            # no index path given, index is in-memory, persist it
            state["_index"] = save_index_to_temp_file(self._index.save, progress=self.parameters.progress_save)
        return state



import faiss
from safecheck import Float, Float32, Int64, NumpyArray, typecheck
from typing_extensions import Any, TypeVar

from ._base import NearestNeighbors
from ._base._helpers import IndexWrapper

__all__ = ["FaissIndex", "FaissNeighbors"]

T = TypeVar("T", bound=type[faiss.Index])


class FaissIndex(IndexWrapper[T]):
    def __init__(self, index: T, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.index = index


class FaissNeighbors(NearestNeighbors):
    """A simple wrapper around ``faiss.Index``.

    A Faiss index is built from any combination of the following.

    1. Vector transform — a pre-processing step applied to vectors before indexing (PCA, OPQ).
    2. Coarse quantizer — rough organization of vectors to subdomains (for restricting search
       scope, includes IVF, IMI, and HNSW).
    3. Fine quantizer — a finer compression of vectors into smaller domains (for compressing
       index size, such as PQ).
    4. Refinement — a final step at search-time which re-orders results using distance calculations
       on the original flat vectors. Alternatively, another index (non-flat) index can be used.

    Note that coarse quantization refers to the 'clustering' of vectors (such as inverted indexing with IVF).
    By using coarse quantization, we enable non-exhaustive search by limiting the search scope.
    Fine quantization describes the compression of vectors into codes (as with PQ). The purpose fine quantization
    is to reduce the memory usage of the index.

    References
    ----------
        - https://github.com/facebookresearch/faiss
        - https://www.pinecone.io/learn/series/faiss/

    """

    @typecheck
    def __init__(
        self,
        *,
        index: str | FaissIndex | faiss.Index = "Flat",
        add_data_on_fit: bool = True,
    ) -> None:
        """Instantiate faiss nearest neighbors.

        :param index: An index factory string, a FaissIndex wrapped index or a faiss.Index.
        :param add_data_on_fit: Add the data used for index training to the learned index.
        """
        super().__init__()
        # to be defined in ``fit``
        self._index: faiss.Index | None = None

    @typecheck
    def fit(self, data: Float[NumpyArray, "n d"]) -> "FaissNeighbors":
        _, dim = data.shape
        self._index = self._create_index(dim)
        self._index.train(data)  # type: ignore[reportCallIssue]

        if self.parameters.add_data_on_fit:
            # data might be added directly on fit, or using the ``add`` method
            self._index.add(data)  # type: ignore[reportCallIssue]

        return self

    @typecheck
    def add(self, data: Float[NumpyArray, "n d"]) -> "FaissNeighbors":
        self._index.add(data)  # type: ignore[reportCallIssue]
        return self

    def query(
        self,
        point: Float[NumpyArray, "d"],
        n_neighbors: int,
    ) -> tuple[Int64[NumpyArray, "{n_neighbors}"], Float32[NumpyArray, "{n_neighbors}"]]:
        idx, dist = self.query_batch(point.reshape(1, -1), n_neighbors)
        return idx.ravel(), dist.ravel()

    def query_batch(
        self,
        points: Float[NumpyArray, "m d"],
        n_neighbors: int,
    ) -> tuple[Int64[NumpyArray, "m {n_neighbors}"], Float32[NumpyArray, "m {n_neighbors}"]]:
        dist, idx = self._index.search(points, n_neighbors)  # type: ignore[reportCallIssue]
        return idx, dist

    def _create_index(self, dim: int) -> faiss.Index:
        index = self.parameters.index
        if isinstance(index, str):
            # if the index is a string, we treat it as an index factory
            return faiss.index_factory(dim, index)

        if isinstance(index, faiss.Index):
            # if the index is an existing index, simply use that
            return index

        # otherwise the index should be a callable returning an index
        return index(dim)
